from __future__ import annotations


def example_func(a: str | None, b: str | int | None = None) -> dict[str, int]:
    """Example function.

    Hover over the parameter and return type annotations to see the long versions.

    Args:
        a: An example parameter
        b: Another, with a default
    Returns:
        An example return value
    """
    pass
