@import "sphinx-book-theme.css";

/* for the sphinx design cards */
body {
    --sd-color-shadow: dimgrey;
}

dl.citation > dt {
    float: left;
    margin-right: 15px;
    font-weight: bold;
}

/* for custom small role */
.small {
    font-size: 40% !important;
}

.smaller,
.pr {
    font-size: 70% !important;
}
